//================================================================================================
// $Id: codegen_enum.cpp 6514 2023-04-03 18:02:07Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2016-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "codegen.h"
#include "tinyframe.h"
#include "cgsection.h"
#include "welltype/wtruntime.h"
#include "wtruntime/types.h"
#include "welltype/stdlib.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_enumtype.h"
#include "wtparser/ast/ast_enumerator.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"
#include <vector>

namespace wtc
{
namespace codegen
{

void CodeGen::GenEnum_succ(wtast::FunctionSignature* sig)
{
	wtast::EnumType* const enum_type = static_cast<wtast::EnumType*>(sig->GetArg(0)->GetType());

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	const std::string lab_error = mEngine.GenLabel();

	mEngine.MakeDUP(enum_type->GetLoadSize());
	mEngine.MakePushImm(wt_u32(enum_type->GetNumberOfEnumerators()));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddJFI(lab_error);

	// then
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);

	// else
	mEngine.AddLabel(lab_error);
	mEngine.MakePOP(enum_type->GetLoadSize());
	mEngine.MakePBA(mExternals->GetExcAddr(ConstraintErrExc));
	mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
	mEngine.AddSingle(WT_TOMB_OP_ERAISE);
}

void CodeGen::GenEnum_pred(wtast::FunctionSignature* sig)
{
	wtast::EnumType* const enum_type = static_cast<wtast::EnumType*>(sig->GetArg(0)->GetType());

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	const std::string lab_error = mEngine.GenLabel();

	mEngine.MakeDUP(enum_type->GetLoadSize());
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 1);
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddJFI(lab_error);

	// then
	mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);

	// else
	mEngine.AddLabel(lab_error);
	mEngine.MakePOP(enum_type->GetLoadSize());
	mEngine.MakePBA(mExternals->GetExcAddr(ConstraintErrExc));
	mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
	mEngine.AddSingle(WT_TOMB_OP_ERAISE);
}

void CodeGen::GenEnum_ordinal(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_tos(wtast::FunctionSignature* sig)
{
	wtast::EnumType* const enum_type = static_cast<wtast::EnumType*>(sig->GetArg(0)->GetType());

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	std::vector<std::string> labs;
	const std::size_t num = enum_type->GetNumberOfEnumerators();

	for(std::size_t i=num;0!=i;--i)
	{
		labs.push_back(mEngine.GenLabel());
	}

	for(std::size_t i=num;1<i;--i)
	{
		mEngine.AddSingle(WT_TOMB_OP_SUB_I, 1);
		mEngine.MakeDUP(enum_type->GetLoadSize());
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
		mEngine.AddSingle(WT_TOMB_OP_CMP); // results { -1, 0, 1 }
		mEngine.AddJFL(labs[num-i]); // jump if CMP was 0 => equality!
	}

	mEngine.AddJmpL(labs[num-1]);

	for(std::size_t i=num;0!=i;--i)
	{
		const std::string & str = enum_type->GetEnumerator(i-1)->GetName();
		const std::size_t   len = str.length();

		mEngine.AddLabel(labs[i-1]);
		mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, ' ');
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, len);
		mEngine.MakeEXEC(WTMP_STRCHARS);

		for(std::size_t j=0;j<len;++j)
		{
			mEngine.AddSingle(WT_TOMB_OP_PWZ_I, wt_u8(str[j]));
			mEngine.AddSingle(WT_TOMB_OP_PDSP_I, sizeof(wt_u32));
			mEngine.AddSingle(WT_TOMB_OP_IDRD_I, 0);
			mEngine.MakeSTRCR(wt_u32(j));
			mEngine.AddSingle(WT_TOMB_OP_STB);
		}

		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
	}
}

void CodeGen::GenEnum_toEnum_u32(wtast::FunctionSignature* sig)
{
	wtast::EnumType* const enum_type = static_cast<wtast::EnumType*>(sig->GetRet(0));

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	const std::string lab_error = mEngine.GenLabel();

	mEngine.MakeDUP(enum_type->GetLoadSize());
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 1);
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QLT);
	mEngine.AddJTL(lab_error);

	mEngine.MakeDUP(enum_type->GetLoadSize());
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, enum_type->GetNumberOfEnumerators());
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGT);
	mEngine.AddJTL(lab_error);

	mEngine.AddSingle(WT_TOMB_OP_LEAVE);

	mEngine.AddLabel(lab_error);
	mEngine.MakePOP(enum_type->GetLoadSize());
	mEngine.MakePBA(mExternals->GetExcAddr(ConstraintErrExc));
	mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
	mEngine.AddSingle(WT_TOMB_OP_ERAISE);
}

void CodeGen::GenEnum_toEnum_string(wtast::FunctionSignature* sig)
{
	wtast::EnumType* const enum_type = static_cast<wtast::EnumType*>(sig->GetRet(0));

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	const std::string lab_error = mEngine.GenLabel();

	typedef std::map<std::size_t, std::set<std::string> > strs_t;
	strs_t strs;

	for(std::size_t i=0,n=enum_type->GetNumberOfEnumerators();i!=n;++i)
	{
		const std::string & e = enum_type->GetEnumerator(i)->GetName();

		strs[e.length()].insert(e);
	}

	mEngine.MakeDUP(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_STRL);

	for(strs_t::const_iterator it=strs.begin(),end=strs.end();it!=end;++it)
	{
		const std::string lab_after = mEngine.GenLabel();

		mEngine.MakeDUP(sizeof(wt_u32));
		mEngine.MakePushImm(wt_u32(it->first));
		mEngine.AddSingle(WT_TOMB_OP_CMP);
		mEngine.AddSingle(WT_TOMB_OP_QEQ);
		mEngine.AddJFL(lab_after);
		mEngine.MakePOP(sizeof(wt_u32));

		GenEnum__toEnum_string__impl(enum_type, std::string(), it->second, lab_error);

		mEngine.AddLabel(lab_after);
	}

	mEngine.MakePOP(sizeof(wt_u32));
	mEngine.AddLabel(lab_error);
	mEngine.MakePOP(sizeof(wt_vmptr));
	mEngine.MakePBA(mExternals->GetExcAddr(ConstraintErrExc));
	mEngine.AddSingle(WT_TOMB_OP_LD_I, sizeof(wt_exception));
	mEngine.AddSingle(WT_TOMB_OP_ERAISE);
}

void CodeGen::GenEnum__toEnum_string__impl(wtast::EnumType* enum_type, std::string prefix, const std::set<std::string> & remain, const std::string & lab_error)
{
	if(remain.empty() || (1==remain.size() && remain.begin()->empty()))
	{
		wtast::Enumerator* enumerator = enum_type->GetEnumerator(prefix);

		if(AddAssert(mProblemHandler, enumerator))
		{
			return;
		}

		mEngine.MakePOP(sizeof(wt_vmptr));
		mEngine.MakePushImm(enumerator->GetOrdinal());
		mEngine.AddSingle(WT_TOMB_OP_LEAVE);
		return;
	}

	typedef std::map<char, std::set<std::string> > next_t;
	next_t next;

	for(std::set<std::string>::const_iterator it=remain.begin(),end=remain.end();it!=end;++it)
	{
		next[(*it)[0]].insert(it->substr(1));
	}

	const wt_u32 char_idx = wt_u32(prefix.length());

	std::size_t num = next.size();
	for(next_t::const_iterator it=next.begin(),end=next.end();it!=end;--num,++it)
	{
		mEngine.MakeDUP(sizeof(wt_vmptr));
		mEngine.MakeSTRCR(char_idx);
		mEngine.AddSingle(WT_TOMB_OP_LDB);
		mEngine.AddSingle(WT_TOMB_OP_PWZ_I, wt_u8(it->first));
		mEngine.AddSingle(WT_TOMB_OP_CMP);
		mEngine.AddSingle(WT_TOMB_OP_QEQ);

		std::string opt_lab_after;

		if(1==num)
		{
			mEngine.AddJFL(lab_error);
		}
		else
		{
			opt_lab_after = mEngine.GenLabel();
			mEngine.AddJFL(opt_lab_after);
		}

		GenEnum__toEnum_string__impl(enum_type, prefix+it->first, it->second, lab_error);

		if(!opt_lab_after.empty())
		{
			mEngine.AddLabel(opt_lab_after);
		}
	}
}

void CodeGen::GenEnum_eq(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QEQ);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_ne(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QNE);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_lt(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QLT);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_le(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QLE);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_gt(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGT);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

void CodeGen::GenEnum_ge(wtast::FunctionSignature* sig)
{
	mEngine.AddGlobal(sig->GetCanonicalName(false, true));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGE);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

} // codegen
} // wtc
