//================================================================================================
// $Id: framestack.h 6505 2023-04-03 17:50:17Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Frame stack
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTC_FRAMESTACK_H
#define WTC_FRAMESTACK_H

#include "welltype/wtruntime.h"
#include "wtparser/uniqueid.h"
#include <string>
#include <map>
#include <vector>

namespace wtast
{
class Type;
class VariableDeclFragment;
class FunctionArgument;
} // namespace wtast

namespace wtc
{

class FrameStack
{
public:
	struct Variable
	{
		Variable() { }

		Variable(const std::string & n, wtast::Type* t, wt_u32 o, wt_u32 d, bool g);

		std::string  name;
		wtast::Type* type;
		wt_u32       offset;
		wt_u32       depth;
		bool         is_global;
	};

	typedef std::map<wtparser::UniqueId, Variable> varmap_t;
	typedef std::vector<Variable*> varlist_t;

private:
	struct Frame
	{
		Frame(const std::string & name, bool is_global=false)
			: name(name)
			, is_global(is_global)
			, have_frag32(false)
			, frag32_offs(0)
			, dvt_size(0)
			, objects_below(0)
		{
			// ok
		}

		Frame(const std::string & name, const Frame & frame)
			: name(name)
			, vars(frame.vars)
			, is_global(frame.is_global)
			, have_frag32(false)
			, frag32_offs(0)
			, dvt_size(frame.dvt_size)
			, objects_below(frame.objects_below + frame.objects.size())
		{
			// ok
		}

		const Variable* Find(wtparser::UniqueId unique_id) const
		{
			varmap_t::const_iterator it = vars.find(unique_id);

			if(it!=vars.end())
			{
				return &it->second;
			}

			return 0;
		}

		void Insert(wtparser::UniqueId unique_id, const std::string & name, bool funcarg, wtast::Type* type, wt_u32 & offs, wt_u32 depth);

		std::string name;
		varmap_t    vars;
		varlist_t   objects; // objects in current frame
		bool        is_global;
		bool        have_frag32;
		wt_u32      frag32_offs;
		wt_u32      dvt_size;
		wt_u32      objects_below;
	};

	typedef std::vector<Frame> framestack_t;

	Frame         mGlobal;
	framestack_t  mFrameStack;
	wt_u32        mTotalDvtSize;
	wt_u32        mTotalCsosSize;
	std::string   mExHanLabel;
	std::string   mFinallyLabel;

public:
	FrameStack();

	~FrameStack()
	{
		// ok
	}

	bool EnterFrame(const std::string & name);
	void LeaveFrame();
	bool NoMoreFrames() const { return mFrameStack.empty(); }
	bool IsLastFrame() const { return 1u==mFrameStack.size(); }
	bool AddVariable(wtast::VariableDeclFragment* frag, bool is_global);
	bool AddVariable(wtast::VariableDeclFragment* frag, bool is_global, wt_u32 & offs);
	bool AddVariable(wtast::FunctionArgument* arg);
	bool AddVariable(wtast::Type* type, const std::string & name, wtparser::UniqueId unique_id);
	bool AddVariable(wtast::Type* type, const std::string & name, wtparser::UniqueId unique_id, wt_u32 & offs);
	wt_u32 CurrentDepth() const { return mFrameStack.size(); }
	wt_u32 GetDvtSize();
	wt_u32 GetCsosSize();
	wt_u32 GetNumObjectsInFrame();
	wt_u32 GetGlobalDvtSize();
	const varlist_t & GetObjectsInFrame() const;
	const std::string & GetExHanLabel() const { return mExHanLabel; }
	const std::string & GetFinallyLabel() const { return mFinallyLabel; }
	const Variable* FindVar(wtparser::UniqueId unique_id) const;
};

} // wtc

#endif // WTC_FRAMESTACK_H
