#include "config_private.h"
#include "hash.h"
#include "byteorder.h"
#include "wtruntime.h"
#include <string.h>

// TODO: make a public symbol from this: WTR_HASH_MAX_DIGEST_LEN
#define WT_HASH_MAX_LEN  (512/8)

const struct wtr_hashalg* wtr_hash_find_by_name(
	const struct wtr_hashalg* algs, const char* const name)
{
	if(NULL==algs)
	{
		return NULL;
	}

	while(NULL!=algs->name)
	{
		if(0==strcmp(name, algs->name))
		{
			return algs;
		}

		++algs;
	}

	return NULL;
}

const struct wtr_hashalg* wtr_hash_find_by_id(
	const struct wtr_hashalg* algs, const wt_u16 id)
{
	if(NULL==algs)
	{
		return NULL;
	}

	while(NULL!=algs->name)
	{
		if(id==algs->hashalg_id)
		{
			return algs;
		}

		++algs;
	}

	return NULL;
}

static wt_error wtr_hash_visitor__(const struct wtr_hashalg* const alg,
	void* const ctx, const void* data, size_t size, wt_bo bo)
{
	const wt_u8* const   base = (const wt_u8*)data;
	const wtb_hdr* const hdr  = (const wtb_hdr*)data;

	if(size<sizeof(wtb_hdr))
	{
		return WTE_BINERR;
	}

	if(WTBMAG0!=hdr->b_ident[WTBI_MAGIC0] ||
		WTBMAG1!=hdr->b_ident[WTBI_MAGIC1] ||
		WTBMAG2!=hdr->b_ident[WTBI_MAGIC2] ||
		WTBMAG3!=hdr->b_ident[WTBI_MAGIC3] ||
		WTBV_CURRENT!=hdr->b_ident[WTBI_VERSION] ||
		bo!=hdr->b_ident[WTBI_BYTEORDER])
	{
		return WTE_BINERR;
	}

	/* header */
	{
		alg->update(alg, ctx, hdr, sizeof(wtb_hdr));
	}

	/* section headers */
	{
		wt_u16 b_shnum = wtr_botoh16(hdr->b_shnum, bo);
		wt_u32 b_shoff = wtr_botoh32(hdr->b_shoff, bo);

		alg->update(alg, ctx, base+b_shoff, sizeof(wtb_shdr)*b_shnum);
	}

	/* thread headers */
	{
		wt_u16 b_thnum = wtr_botoh16(hdr->b_thnum, bo);
		wt_u32 b_thoff = wtr_botoh32(hdr->b_thoff, bo);

		alg->update(alg, ctx, base+b_thoff, sizeof(wtb_thdr)*b_thnum);
	}

	/* section data */
	{
		wt_u16 i;
		wt_u16 b_shnum = wtr_botoh16(hdr->b_shnum, bo);
		wt_u32 b_shoff = wtr_botoh32(hdr->b_shoff, bo);

		const wtb_shdr* shdr = (const wtb_shdr*)(base + b_shoff);

		for(i=1;i<b_shnum;++shdr,++i)
		{
			wt_u32 sh_size = wtr_botoh32(shdr->sh_size, bo);
			wt_u32 sh_offs = wtr_botoh32(shdr->sh_offs, bo);

			/* we don't want to hash 0 length data, because it does not any sense */
			if(0==sh_size)
			{
				continue;
			}

			/* CLEAR and LOAD flags excludes themselves */
			if((WTBSF_CLEAR & shdr->sh_flag) && (WTBSF_LOAD & shdr->sh_flag))
			{
				return WTE_BINERR;
			}

			/* if the CLEAR flag is set, it means no associated data */
			if(WTBSF_CLEAR & shdr->sh_flag)
			{
				continue;
			}

			/* INTEGRITY sections: we can hash only the headers, and we should do it */
			/* -- otherwise problems will raise to find the fix-point of the hash value */
			if(WTBST_INTEGRITY==shdr->sh_type)
			{
				alg->update(alg, ctx, base+sh_offs, sizeof(wtb_integhdr));
				continue;
			}

			/* section data is ready to be hashed */
			alg->update(alg, ctx, base+sh_offs, sh_size);
		}
	}

	return WTE_OK;
}

wt_error wtr_hash_calc(void* out_digest, const struct wtr_hashalg* const alg,
	const void* data, size_t size, wt_bo bo)
{
	wt_error err  = WTE_OK;

	void* ctx;

	if(0!=alg->init(alg, &ctx))
	{
		// TODO: maybe a better error code?
		return WTE_INTEG;
	}

	if(WTE_OK!=(err=wtr_hash_visitor__(alg, ctx, data, size, bo)))
	{
		return err;
	}

	alg->final(alg, ctx, out_digest);

	return WTE_OK;
}

wt_error wtr_hash_check(const void* digest,
	const struct wtr_hashalg* const alg, const void* data, size_t size,
	wt_bo bo)
{
	wt_u8    calc_digest[WT_HASH_MAX_LEN];
	wt_error err;

	if(0==alg->digest_len || WT_HASH_MAX_LEN<alg->digest_len)
	{
		// TODO: maybe a better error code?
		return WTE_INTEG;
	}

	if(WTE_OK!=(err=wtr_hash_calc(calc_digest, alg, data, size, bo)))
	{
		return err;
	}

	if(0!=memcmp(digest, calc_digest, alg->digest_len))
	{
		return WTE_INTEG;
	}

	return WTE_OK;
}
