#include "config_private.h"
#include "microprog.h"
#include "wtruntime.h"
#include "version.h"
#include "vm.h"
#include "program.h"
#include "ctx.h"
#include "processor.h"
#include "objects.h"
#include "stdlib.h"
#include "convert.h"
#include "byteorder.h"
#include "iocall.h"
#include <string.h>
#include <ctype.h>
#if WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64
#  include <math.h>
#  include <float.h>
#endif /* WTRUNTIME_ENABLE_FLOAT32 || WTRUNTIME_ENABLE_FLOAT64 */

#define __WTR_LIST_MICROPROGRAM(lowercase,UPPERCASE) \
	__WTR_NAME_MICROPROGRAM(lowercase,UPPERCASE),

#if !WTRUNTIME_ENABLE_FLOAT32 || !WTRUNTIME_ENABLE_FLOAT64
#  define wtr_mp_cf32f64        wtr_mp_undefined
#  define wtr_mp_cf64f32        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT32 || !WTRUNTIME_ENABLE_FLOAT64 */
#if !WTRUNTIME_ENABLE_FLOAT32
#  define wtr_mp_randf          wtr_mp_undefined
#  define wtr_mp_f32sqrt        wtr_mp_undefined
#  define wtr_mp_f32log         wtr_mp_undefined
#  define wtr_mp_f32exp         wtr_mp_undefined
#  define wtr_mp_f32sin         wtr_mp_undefined
#  define wtr_mp_f32cos         wtr_mp_undefined
#  define wtr_mp_f32tan         wtr_mp_undefined
#  define wtr_mp_f32asin        wtr_mp_undefined
#  define wtr_mp_f32acos        wtr_mp_undefined
#  define wtr_mp_f32atan        wtr_mp_undefined
#  define wtr_mp_f32atan2       wtr_mp_undefined
#  define wtr_mp_f32pow         wtr_mp_undefined
#  define wtr_mp_cf32string     wtr_mp_undefined
#  define wtr_mp_cstringf32     wtr_mp_undefined
#  define wtr_mp_cu32f32        wtr_mp_undefined
#  define wtr_mp_ci32f32        wtr_mp_undefined
#  define wtr_mp_cf32i32        wtr_mp_undefined
#  define wtr_mp_cf32u32        wtr_mp_undefined
#  define wtr_mp_ci64f32        wtr_mp_undefined
#  define wtr_mp_cu64f32        wtr_mp_undefined
#  define wtr_mp_cf32i64        wtr_mp_undefined
#  define wtr_mp_cf32u64        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT32 */
#if !WTRUNTIME_ENABLE_FLOAT64
#  define wtr_mp_f64sqrt        wtr_mp_undefined
#  define wtr_mp_f64log         wtr_mp_undefined
#  define wtr_mp_f64exp         wtr_mp_undefined
#  define wtr_mp_f64sin         wtr_mp_undefined
#  define wtr_mp_f64cos         wtr_mp_undefined
#  define wtr_mp_f64tan         wtr_mp_undefined
#  define wtr_mp_f64asin        wtr_mp_undefined
#  define wtr_mp_f64acos        wtr_mp_undefined
#  define wtr_mp_f64atan        wtr_mp_undefined
#  define wtr_mp_f64atan2       wtr_mp_undefined
#  define wtr_mp_f64pow         wtr_mp_undefined
#  define wtr_mp_cf64string     wtr_mp_undefined
#  define wtr_mp_cstringf64     wtr_mp_undefined
#  define wtr_mp_cu32f64        wtr_mp_undefined
#  define wtr_mp_ci32f64        wtr_mp_undefined
#  define wtr_mp_cf64i32        wtr_mp_undefined
#  define wtr_mp_cf64u32        wtr_mp_undefined
#  define wtr_mp_ci64f64        wtr_mp_undefined
#  define wtr_mp_cu64f64        wtr_mp_undefined
#  define wtr_mp_cf64i64        wtr_mp_undefined
#  define wtr_mp_cf64u64        wtr_mp_undefined
#endif /* !WTRUNTIME_ENABLE_FLOAT64 */

const WTR_PFN_MICROPROGRAM wtr_microprogtab[] =
{
	WTR_MICROPROGRAMS(__WTR_LIST_MICROPROGRAM)
};

#define CHECK_POP(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SIZE(_ctx_)>=(_size_)) \
		{ \
			/* branch prediction */ \
		} \
		else \
		{ \
			(_ctx_)->last_exception = StackOverflowException; \
			return WTRR_RAISE; \
		} \
	} \
	while(0)

#define CHECK_PUSH(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SPACE(_ctx_)>=(_size_)) \
		{ \
			/* branch prediction */ \
		} \
		else \
		{ \
			(_ctx_)->last_exception = StackOverflowException; \
			return WTRR_RAISE; \
		} \
	} \
	while(0)

static enum wtr_runresult str_push_and_register__tail__(
	struct wtr_vm* const vm, struct wtr_context* const ctx, wt_vmptr str)
{
	struct wtr_objentry ent;

	WTR_CTX_DS_PUSH(ctx, wt_vmptr, str);

	ent.obj     = str;
	ent.pfn.ptr = ctx->frame.prog;
	ent.pfn.ext = vm->pc_strd_leave;

	if(NULL==wtr_dynstack_push_p(&ctx->tos, &ent))
	{
		wtr_str_delete(vm, ent.obj);
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

static const wt_u8 _squares_u8[] =
{
	  0,   1,   4,   9,
	 16,  25,  36,  49,
	 64,  81, 100, 121,
	144, 169, 196, 225,
};

static wt_u8 sqrt_u8(wt_u8 x)
{
	const wt_u8* p = _squares_u8;

	if(p[8] <= x) { p += 8; }
	if(p[4] <= x) { p += 4; }
	if(p[2] <= x) { p += 2; }
	if(p[1] <= x) { p += 1; }

	return (wt_u8)(p - _squares_u8);
}

static wt_i8 sqrt_i8(wt_i8 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u8(n);
}

static wt_u32 sqrt_u32(wt_u32 remainder)
{
	wt_u32 place = ((wt_u32)1) << (sizeof(wt_u32) * 8 - 2); /* calculated by compiler = same runtime as: place = 0x40000000 */
	wt_u32 root  = 0;

	while(place>remainder)
	{
		place >>= 2;
	}

	while(place)
	{
		if(remainder>=root+place)
		{
			remainder -= root+place;
			root += place << 1;
		}

		root >>= 1;
		place >>= 2;
	}

	return root;
}

static wt_i32 sqrt_i32(wt_i32 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u32((wt_u32)n);
}

static wt_i16 sqrt_i16(wt_i16 n)
{
	if(0>=n)
	{
		return 0;
	}

	return sqrt_u32((wt_u32)(wt_u16)n);
}

static wt_u16 sqrt_u16(wt_u16 n)
{
	return sqrt_u32((wt_u32)n);
}

static wt_i64 sqrt_i64(wt_i64 n0)
{
	wt_u64 n = n0, m, y, b;

	if(n0<=0) return 0;

	m = 0x4000000000000000ull; /* TODO: UINT64_C */
	y = 0ul;

	while(m != 0ul)
	{
		b = y | m;
		y = y >> 1;
		if(n >= b)
		{
			n = n - b;
			y = y | m;
		}
		m = m >> 2;
	}
	return y;
}

static wt_u64 sqrt_u64(wt_u64 n)
{
	wt_u64 m, y, b;

	m = 0x4000000000000000ull; /* TODO: UINT64_C */
	y = 0ul;

	while(m != 0ul)
	{
		b = y | m;
		y = y >> 1;
		if(n >= b)
		{
			n = n - b;
			y = y | m;
		}
		m = m >> 2;
	}
	return y;
}

static wt_u32 _wtr_xorshift(struct wtr_vm* const vm)
{
	wt_u32 t = vm->xs_x ^ (vm->xs_x << 11);
	vm->xs_x = vm->xs_y; vm->xs_y = vm->xs_z; vm->xs_z = vm->xs_w;
	return vm->xs_w = vm->xs_w ^ (vm->xs_w >> 19) ^ t ^ (t >> 8);
}

enum wtr_runresult wtr_mp_undefined(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	((void)vm);

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_conwrite(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);
	ASSERT_OBJECT_EXISTENCE(str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_writestdout(ctx, str+offsetof(struct wtr_str, data), len);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_conwritech(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	*ctx->sbuf = (char)ch;

	wtr_iocall_writestdout_x(ctx, ctx->sbuf, 1);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_conreadln(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	wtr_iocall_readlnstdin(ctx, (char)ch);
	return WTRR_IOCALL;
}

/* generic unary, i8 */
static enum wtr_runresult _wtr_mp__g_u_i8(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_i8 (*func)(wt_i8 val))
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_i32, func((wt_i8)(wt_u8)val));

	return WTRR_RUN;
}

/* generic unary, u8 */
static enum wtr_runresult _wtr_mp__g_u_u8(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_u8 (*func)(wt_u8 val))
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, func((wt_u8)val));

	return WTRR_RUN;
}

/* generic unary, i16 */
static enum wtr_runresult _wtr_mp__g_u_i16(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_i16 (*func)(wt_i16 val))
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_i32, func((wt_i16)(wt_u16)val));

	return WTRR_RUN;
}

/* generic unary, u16 */
static enum wtr_runresult _wtr_mp__g_u_u16(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_u16 (*func)(wt_u16 val))
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, func((wt_u16)val));

	return WTRR_RUN;
}

/* generic unary, i32 */
static enum wtr_runresult _wtr_mp__g_u_i32(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_i32 (*func)(wt_i32 val))
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_i32, func(val));

	return WTRR_RUN;
}

/* generic unary, u32 */
static enum wtr_runresult _wtr_mp__g_u_u32(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_u32 (*func)(wt_u32 val))
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, func(val));

	return WTRR_RUN;
}

/* generic unary, i64 */
static enum wtr_runresult _wtr_mp__g_u_i64(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_i64 (*func)(wt_i64 val))
{
	wt_i64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i64, val);
	WTR_CTX_DS_PUSH(ctx, wt_i64, func(val));

	return WTRR_RUN;
}

/* generic unary, u64 */
static enum wtr_runresult _wtr_mp__g_u_u64(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_u64 (*func)(wt_u64 val))
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_u64, func(val));

	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

/* generic unary, f32 */
static enum wtr_runresult _wtr_mp__g_u_f32(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_f32 (*func)(wt_f32 val))
{
	wt_f32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_f32, val);
	WTR_CTX_DS_PUSH(ctx, wt_f32, func(val));

	return WTRR_RUN;
}

/* generic binary, f32 */
static enum wtr_runresult _wtr_mp__g_b_f32(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_f32 (*func)(wt_f32 lhs, wt_f32 rhs))
{
	wt_f32 lhs;
	wt_f32 rhs;

	CHECK_POP(ctx, sizeof(wt_f32)*2);
	WTR_CTX_DS_POP(ctx, wt_f32, rhs);
	WTR_CTX_DS_POP(ctx, wt_f32, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_f32, func(lhs, rhs));

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

/* generic unary, f64 */
static enum wtr_runresult _wtr_mp__g_u_f64(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_f64 (*func)(wt_f64))
{
	wt_f64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_f64, val);
	WTR_CTX_DS_PUSH(ctx, wt_f64, func(val));

	return WTRR_RUN;
}

/* generic binary, f64 */
static enum wtr_runresult _wtr_mp__g_b_f64(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_f64 (*func)(wt_f64 lhs, wt_f64 rhs))
{
	wt_f64 lhs;
	wt_f64 rhs;

	CHECK_POP(ctx, sizeof(wt_f64)*2);
	WTR_CTX_DS_POP(ctx, wt_f64, rhs);
	WTR_CTX_DS_POP(ctx, wt_f64, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_f64, func(lhs, rhs));

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

enum wtr_runresult wtr_mp_i8sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_i8(vm, ctx, sqrt_i8);
}

enum wtr_runresult wtr_mp_u8sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u8(vm, ctx, sqrt_u8);
}

enum wtr_runresult wtr_mp_i16sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_i16(vm, ctx, sqrt_i16);
}

enum wtr_runresult wtr_mp_u16sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u16(vm, ctx, sqrt_u16);
}

enum wtr_runresult wtr_mp_i32sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_i32(vm, ctx, sqrt_i32);
}

enum wtr_runresult wtr_mp_u32sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, sqrt_u32);
}

enum wtr_runresult wtr_mp_i64sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_i64(vm, ctx, sqrt_i64);
}

enum wtr_runresult wtr_mp_u64sqrt(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u64(vm, ctx, sqrt_u64);
}

#define WRAP_UNOP(_type_, _func_) \
static wt_##_type_ _wrap_##_func_##_##_type_(wt_##_type_ val) \
{ \
	return (wt_##_type_)_func_(val); \
} \
enum wtr_runresult wtr_mp_##_type_##_func_(struct wtr_vm* const vm, struct wtr_context* const ctx) \
{ \
	return _wtr_mp__g_u_##_type_(vm, ctx, _wrap_##_func_##_##_type_); \
} \

#define WRAP_BINOP(_type_, _func_) \
static wt_##_type_ _wrap_##_func_##_##_type_(wt_##_type_ lhs, wt_##_type_ rhs) \
{ \
	return (wt_##_type_)_func_(lhs, rhs); \
} \
enum wtr_runresult wtr_mp_##_type_##_func_(struct wtr_vm* const vm, struct wtr_context* const ctx) \
{ \
	return _wtr_mp__g_b_##_type_(vm, ctx, _wrap_##_func_##_##_type_); \
} \

#if WTRUNTIME_ENABLE_FLOAT32

WRAP_UNOP(f32, sqrt)
WRAP_UNOP(f32, log)
WRAP_UNOP(f32, exp)
WRAP_UNOP(f32, sin)
WRAP_UNOP(f32, cos)
WRAP_UNOP(f32, tan)
WRAP_UNOP(f32, asin)
WRAP_UNOP(f32, acos)
WRAP_UNOP(f32, atan)
WRAP_BINOP(f32, atan2)
WRAP_BINOP(f32, pow)

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

WRAP_UNOP(f64, sqrt)
WRAP_UNOP(f64, log)
WRAP_UNOP(f64, exp)
WRAP_UNOP(f64, sin)
WRAP_UNOP(f64, cos)
WRAP_UNOP(f64, tan)
WRAP_UNOP(f64, asin)
WRAP_UNOP(f64, acos)
WRAP_UNOP(f64, atan)
WRAP_BINOP(f64, atan2)
WRAP_BINOP(f64, pow)

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

enum wtr_runresult wtr_mp_sarb(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 amount;
	wt_u32 value;

	CHECK_POP(ctx, sizeof(value)*2);
	WTR_CTX_DS_POP(ctx, wt_u32, amount);
	WTR_CTX_DS_POP(ctx, wt_u32, value);

	if(amount<=7)
	{
		const wt_u32 shifted_up   = value << 24;
		const wt_i32 shifted_down = ((wt_i32)shifted_up) >> (24 + amount);
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)shifted_down) & 0xFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_sarh(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 amount;
	wt_u32 value;

	CHECK_POP(ctx, sizeof(value)*2);
	WTR_CTX_DS_POP(ctx, wt_u32, amount);
	WTR_CTX_DS_POP(ctx, wt_u32, value);

	if(amount<=15)
	{
		const wt_u32 shifted_up   = value << 16;
		const wt_i32 shifted_down = ((wt_i32)shifted_up) >> (16 + amount);
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)shifted_down) & 0xFFFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

static wt_u32 _wrap_negb(wt_u32 val)
{
	return (-((wt_i8)val)) & 0xFF;
}

enum wtr_runresult wtr_mp_negb(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_negb);
}

static wt_u32 _wrap_negh(wt_u32 val)
{
	return (-((wt_i16)val)) & 0xFFFF;
}

enum wtr_runresult wtr_mp_negh(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_negh);
}

static enum wtr_runresult _wtr_mp_strcmp_xcase(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const int insensitive)
{
	wt_vmptr lhs;
	wt_vmptr rhs;
	wt_i32   res;
	wt_u32   len_r;
	wt_u32   len_l;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, rhs);
	WTR_CTX_DS_POP(ctx, wt_vmptr, lhs);

	if(0!=wtr_vma_read(vm, rhs+offsetof(struct wtr_str, length), &len_r,
		sizeof(len_r)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, lhs+offsetof(struct wtr_str, length), &len_l,
		sizeof(len_l)))
	{
		goto mav;
	}

	if(len_l<len_r)
	{
		res = -256;
	}
	else
	if(len_l>len_r)
	{
		res = +256;
	}
	else
	{
		res = 0;

		lhs += offsetof(struct wtr_str, data);
		rhs += offsetof(struct wtr_str, data);

		while(0<(len_r--))
		{
			wt_u8 ch_l;
			wt_u8 ch_r;

			if(0!=wtr_vma_read(vm, lhs++, &ch_l, sizeof(ch_l)))
			{
				goto mav;
			}

			if(0!=wtr_vma_read(vm, rhs++, &ch_r, sizeof(ch_r)))
			{
				goto mav;
			}

			if(insensitive)
			{
				ch_l = tolower(ch_l);
				ch_r = tolower(ch_r);
			}

			if(ch_l!=ch_r)
			{
				res = ch_l - ch_r;
				break;
			}
		}
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, res);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strcmp(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_strcmp_xcase(vm, ctx, 0);
}

enum wtr_runresult wtr_mp_stricmp(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_strcmp_xcase(vm, ctx, 1);
}

enum wtr_runresult wtr_mp_strdup(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr src;
	wt_vmptr res;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=(res=wtr_str_create_n(vm, NULL, slen)))
	{
		if(0!=wtr_vma_memcpy(vm,
			res+offsetof(struct wtr_str, data),
			src+offsetof(struct wtr_str, data),
			slen))
		{
			wtr_str_delete(vm, res);
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strcat(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr vma_r;
	wt_vmptr vma_l;
	wt_vmptr res;
	wt_vmptr p;
	wt_u32   len_r;
	wt_u32   len_l;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, vma_r);
	WTR_CTX_DS_POP(ctx, wt_vmptr, vma_l);

	if(0!=wtr_vma_read(vm, vma_r+offsetof(struct wtr_str, length), &len_r,
		sizeof(len_r)) ||
		0!=wtr_vma_read(vm, vma_l+offsetof(struct wtr_str, length), &len_l,
		sizeof(len_l)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0==(res=wtr_str_create_n(vm, NULL, len_l+len_r)))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	p = res + offsetof(struct wtr_str, data);

	if(0!=wtr_vma_memcpy(vm, p, vma_l+offsetof(struct wtr_str, data), len_l)
		||
		0!=wtr_vma_memcpy(vm, p+len_l, vma_r+offsetof(struct wtr_str, data),
		len_r))
	{
		wtr_str_delete(vm, res);
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strrev(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr src;
	wt_vmptr res;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=(res=wtr_str_create_n(vm, NULL, slen)))
	{
		wt_vmptr dst = res+offsetof(struct wtr_str, data);
		wt_vmptr end = dst + slen;

		src += offsetof(struct wtr_str, data)+slen-1;

		while(dst<end)
		{
			if(0!=wtr_vma_memcpy(vm, dst++, src--, 1))
			{
				wtr_str_delete(vm, res);
				ctx->last_exception = MemoryAccessViolationException;
				return WTRR_RAISE;
			}
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strsub(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr src;
	wt_vmptr res;
	wt_u32   slen;
	wt_u32   offset;
	wt_u32   count;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, count);
	WTR_CTX_DS_POP(ctx, wt_u32, offset);
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(offset>=slen)
	{
		offset = slen;
		count  = 0;
	}
	else
	if((offset+count)>slen)
	{
		count  = slen - offset;
	}

	if(0!=(res=wtr_str_create_n(vm, NULL, count)))
	{
		if(0!=wtr_vma_memcpy(vm,
			res+offsetof(struct wtr_str, data),
			src+offsetof(struct wtr_str, data)+offset,
			count))
		{
			wtr_str_delete(vm, res);
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strsub2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr src;
	wt_vmptr res;
	wt_u32   slen;
	wt_u32   offset;
	wt_u32   count;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, offset);
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	count = slen - offset;

	if(offset>=slen)
	{
		offset = slen;
		count  = 0;
	}

	if(0!=(res=wtr_str_create_n(vm, NULL, count)))
	{
		if(0!=wtr_vma_memcpy(vm,
			res+offsetof(struct wtr_str, data),
			src+offsetof(struct wtr_str, data)+offset,
			count))
		{
			wtr_str_delete(vm, res);
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strlen(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, len);

	return WTRR_RUN;
}

static enum wtr_runresult _wtr_mp_upper_lower(struct wtr_vm* const vm,
	struct wtr_context* const ctx, int (*const func)(int))
{
	wt_vmptr src;
	wt_vmptr res;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=(res=wtr_str_create_n(vm, NULL, slen)))
	{
		wt_vmptr dst = res+offsetof(struct wtr_str, data);
		wt_vmptr end = dst + slen;

		src += offsetof(struct wtr_str, data);

		while(dst<end)
		{
			char ch;

			if(0!=wtr_vma_read(vm, src++, &ch, 1))
			{
			delete_and_mav:;
				wtr_str_delete(vm, res);
				ctx->last_exception = MemoryAccessViolationException;
				return WTRR_RAISE;
			}

			ch = func(ch);

			if(0!=wtr_vma_write(vm, dst++, &ch, 1))
			{
				goto delete_and_mav;
			}
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strupper(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_upper_lower(vm, ctx, toupper);
}

enum wtr_runresult wtr_mp_strlower(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_upper_lower(vm, ctx, tolower);
}

struct cset
{
	wt_u32 set[256/(8*sizeof(wt_u32))];
};

static void cset_set(struct cset* const cset, wt_u8 ch)
{
	cset->set[ch/(8*sizeof(wt_u32))] |=
		((wt_u32)1) << (ch % (8*sizeof(wt_u32)));
}

static int cset_is_set(const struct cset* const cset, wt_u8 ch)
{
	return cset->set[ch/(8*sizeof(wt_u32))] &
		((wt_u32)1) << (ch % (8*sizeof(wt_u32)));
}

static int cset_is_not_set(const struct cset* const cset, wt_u8 ch)
{
	return !cset_is_set(cset, ch);
}

static enum wtr_runresult _wtr_mp_strtrim_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const struct cset* const cset)
{
	wt_vmptr src;
	wt_vmptr end;
	wt_vmptr res;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, src);

	if(0!=wtr_vma_read(vm, src+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	src += offsetof(struct wtr_str, data);
	end = src + slen;

	while(src<end)
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, src, &ch, 1))
		{
			goto mav;
		}

		if(!cset_is_set(cset, ch))
		{
			break;
		}

		++src;
	}

	while(src<end)
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, end-1, &ch, 1))
		{
			goto mav;
		}

		if(!cset_is_set(cset, ch))
		{
			break;
		}

		--end;
	}

	slen = end - src;

	if(0!=(res=wtr_str_create_n(vm, NULL, slen)))
	{
		if(0!=wtr_vma_memcpy(vm, res+offsetof(struct wtr_str, data), src,
			slen))
		{
			wtr_str_delete(vm, res);
		mav:;
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strtrim(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	/* 0x00..0x20 */
	static const struct cset ascii_ctrs = { 0xFFFFFFFF, 1, 0, 0, 0, 0, 0, 0 };
	return _wtr_mp_strtrim_impl(vm, ctx, &ascii_ctrs);
}

enum wtr_runresult wtr_mp_strtrim2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr    str;
	wt_u32      slen;
	struct cset cset;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	memset(&cset, 0, sizeof(cset));

	str += offsetof(struct wtr_str, data);
	while(0<(slen--))
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, str++, &ch, 1))
		{
			goto mav;
		}

		cset_set(&cset, ch);
	}

	return _wtr_mp_strtrim_impl(vm, ctx, &cset);
}

enum wtr_runresult wtr_mp_strempty(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, 0==len);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strstartswith(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr lhs;
	wt_vmptr rhs;
	wt_u32   llen;
	wt_u32   rlen;
	wt_bool  result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, rhs);
	WTR_CTX_DS_POP(ctx, wt_vmptr, lhs);

	if(0!=wtr_vma_read(vm, rhs+offsetof(struct wtr_str, length), &rlen,
		sizeof(rlen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, lhs+offsetof(struct wtr_str, length), &llen,
		sizeof(llen)))
	{
		goto mav;
	}

	if(llen<rlen)
	{
		result = WT_FALSE;
	}
	else
	{
		result = (0==wtr_vma_memcmp(vm,
			lhs+offsetof(struct wtr_str, data),
			rhs+offsetof(struct wtr_str, data),
			rlen));
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strendswith(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr lhs;
	wt_vmptr rhs;
	wt_u32   llen;
	wt_u32   rlen;
	wt_bool  result;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, rhs);
	WTR_CTX_DS_POP(ctx, wt_vmptr, lhs);

	if(0!=wtr_vma_read(vm, rhs+offsetof(struct wtr_str, length), &rlen,
		sizeof(rlen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, lhs+offsetof(struct wtr_str, length), &llen,
		sizeof(llen)))
	{
		goto mav;
	}

	if(llen<rlen)
	{
		result = WT_FALSE;
	}
	else
	{
		result = (0==wtr_vma_memcmp(vm,
			lhs+offsetof(struct wtr_str, data)+llen-rlen,
			rhs+offsetof(struct wtr_str, data),
			rlen));
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, result);

	return WTRR_RUN;
}

static enum wtr_runresult _wtr_mp_strchars_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const wt_vmptr suf_vma,
	const wt_u32 suf_len)
{
	wt_char  ch;
	wt_vmptr res;
	wt_u32   n;

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, n);
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	if(0!=(res=wtr_str_create_n(vm, NULL, n+suf_len)))
	{
		wt_vmptr dst = res+offsetof(struct wtr_str, data);

		if(0!=wtr_vma_memset(vm, dst, (wt_u8)ch, n))
		{
			goto delete_and_mav;
		}

		if(0!=suf_len && 0!=wtr_vma_memcpy(vm, dst+n, suf_vma, suf_len))
		{
		delete_and_mav:;
			wtr_str_delete(vm, res);
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strchars(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_strchars_impl(vm, ctx, 0, 0);
}

enum wtr_runresult wtr_mp_strchars2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr suf;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32) + sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, suf);

	if(0!=wtr_vma_read(vm, suf+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	return _wtr_mp_strchars_impl(vm, ctx,
		suf+offsetof(struct wtr_str, data), slen);
}

static enum wtr_runresult _wtr_mp_strcopies_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const wt_vmptr suf_vma,
	const wt_u32 suf_len)
{
	wt_vmptr str;
	wt_vmptr res;
	wt_u32   n;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_vmptr) + sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, n);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	str += offsetof(struct wtr_str, data);

	if(0!=(res=wtr_str_create_n(vm, NULL, n*slen+suf_len)))
	{
		wt_vmptr dst = res+offsetof(struct wtr_str, data);

		while(0<(n--))
		{
			if(0!=wtr_vma_memcpy(vm, dst, str, slen))
			{
				goto delete_and_mav;
			}

			dst += slen;
		}

		if(0!=suf_len && 0!=wtr_vma_memcpy(vm, dst, suf_vma, suf_len))
		{
		delete_and_mav:;
			wtr_str_delete(vm, res);
			ctx->last_exception = MemoryAccessViolationException;
			return WTRR_RAISE;
		}
	}
	else
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, res);
}

enum wtr_runresult wtr_mp_strcopies(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_strcopies_impl(vm, ctx, 0, 0);
}

enum wtr_runresult wtr_mp_strcopies2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr suf;
	wt_u32   slen;

	CHECK_POP(ctx, sizeof(wt_char) + sizeof(wt_u32) + sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, suf);

	if(0!=wtr_vma_read(vm, suf+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	return _wtr_mp_strcopies_impl(vm, ctx,
		suf+offsetof(struct wtr_str, data), slen);
}

enum wtr_runresult wtr_mp_strcpy(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr dstr; /* dest string */
	wt_vmptr sstr; /* src string */
	wt_u32   doff; /* dest offs */
	wt_u32   soff; /* src offs */
	wt_u32   len;  /* requested len */
	wt_u32   dlen; /* dest string len */
	wt_u32   slen; /* src string len */

	CHECK_POP(ctx, sizeof(wt_vmptr)*2 + sizeof(wt_u32)*3);

	WTR_CTX_DS_POP(ctx, wt_u32, len);
	WTR_CTX_DS_POP(ctx, wt_u32, soff);
	WTR_CTX_DS_POP(ctx, wt_vmptr, sstr);
	WTR_CTX_DS_POP(ctx, wt_u32, doff);
	WTR_CTX_DS_POP(ctx, wt_vmptr, dstr);

	if(0!=wtr_vma_read(vm, dstr+offsetof(struct wtr_str, length), &dlen,
		sizeof(dlen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, sstr+offsetof(struct wtr_str, length), &slen,
		sizeof(slen)))
	{
		goto mav;
	}

	if(len>slen || len>dlen || (doff+len)>dlen || (soff+len)>slen)
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_memmove(vm,
		dstr+offsetof(struct wtr_str, data)+doff,
		sstr+offsetof(struct wtr_str, data)+soff,
		len))
	{
		goto mav;
	}

	return WTRR_RUN;
}

static enum wtr_runresult _wtr_mp_strstr_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr needle, wt_u32 offs)
{
	wt_vmptr haystack;
	wt_vmptr base;
	wt_u32   hlen;
	wt_u32   nlen;
	wt_u32   result;

	/* we already checked the pops */
	WTR_CTX_DS_POP(ctx, wt_vmptr, haystack);

	if(0!=wtr_vma_read(vm, haystack+offsetof(struct wtr_str, length), &hlen,
		sizeof(hlen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, needle+offsetof(struct wtr_str, length), &nlen,
		sizeof(nlen)))
	{
		goto mav;
	}

	haystack += offsetof(struct wtr_str, data);
	needle   += offsetof(struct wtr_str, data);

	base   = haystack;
	result = hlen;

	if(offs<=hlen)
	{
		haystack += offs;
		hlen     -= offs;
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	if(hlen>=nlen)
	{
		/* the loop would go 0<=, but it is unsigned, so compensate here */
		hlen -= (nlen-1);
	}
	else
	{
		hlen = 0;
	}

	while(0<(hlen--))
	{
		if(0==wtr_vma_memcmp(vm, haystack, needle, nlen))
		{
			result = haystack - base;
			break;
		}

		++haystack;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strstr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr needle;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, needle);

	return _wtr_mp_strstr_impl(vm, ctx, needle, 0);
}

enum wtr_runresult wtr_mp_strstr2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr needle;
	wt_u32   offs;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_vmptr, needle);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strstr_impl(vm, ctx, needle, offs);
}

static enum wtr_runresult _wtr_mp_strrstr_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_vmptr needle, wt_u32 offs)
{
	wt_vmptr haystack;
	wt_vmptr base;
	wt_u32   hlen;
	wt_u32   nlen;
	wt_u32   result;

	/* we already checked the pops */
	WTR_CTX_DS_POP(ctx, wt_vmptr, haystack);

	if(0!=wtr_vma_read(vm, haystack+offsetof(struct wtr_str, length), &hlen,
		sizeof(hlen)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if(0!=wtr_vma_read(vm, needle+offsetof(struct wtr_str, length), &nlen,
		sizeof(nlen)))
	{
		goto mav;
	}

	haystack += offsetof(struct wtr_str, data);
	needle   += offsetof(struct wtr_str, data);

	base   = haystack;
	result = hlen;

	if(offs<=hlen)
	{
		haystack += offs;
		hlen     -= offs;
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	/* if hlen<nlen, the resultant address is invalid, be the if below will */
	/* deal with this */
	haystack += hlen - nlen;

	if(hlen>=nlen)
	{
		/* the loop would go 0<=, but it is unsigned, so compensate here */
		hlen -= (nlen-1);
	}
	else
	{
		hlen = 0;
	}

	while(0<(hlen--))
	{
		if(0==wtr_vma_memcmp(vm, haystack, needle, nlen))
		{
			result = haystack - base;
			break;
		}

		--haystack;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, result);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrstr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr needle;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, needle);

	return _wtr_mp_strrstr_impl(vm, ctx, needle, 0);
}

enum wtr_runresult wtr_mp_strrstr2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr needle;
	wt_u32   offs;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_vmptr, needle);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strrstr_impl(vm, ctx, needle, offs);
}

static enum wtr_runresult _wtr_mp_strchr_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const wt_u8 looking_for, wt_u32 offs)
{
	wt_vmptr str;
	wt_vmptr base;
	wt_u32   len;

	/* we already checked the pops */
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	str += offsetof(struct wtr_str, data);
	base = str;

	if(offs<=len)
	{
		str += offs;
		len -= offs;
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	while(0<(len--))
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, str, &ch, 1))
		{
			goto mav;
		}

		if(looking_for==ch)
		{
			break;
		}

		++str;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, str-base);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strchr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	return _wtr_mp_strchr_impl(vm, ctx, (wt_u8)ch, 0);
}

enum wtr_runresult wtr_mp_strchr2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;
	wt_u32  offs;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_char, ch);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strchr_impl(vm, ctx, (wt_u8)ch, offs);
}

static enum wtr_runresult _wtr_mp_strrchr_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const wt_u8 looking_for, wt_u32 offs)
{
	wt_vmptr str;
	wt_vmptr base;
	wt_vmptr found;
	wt_u32   len;

	/* we already checked the pops */
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	str += offsetof(struct wtr_str, data);
	base = str;
	found = str + len;

	if(offs<=len)
	{
		str += offs;
		len -= offs;
	}
	else
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	while(0<(len--))
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, str, &ch, 1))
		{
			goto mav;
		}

		if(looking_for==ch)
		{
			found = str;
		}

		++str;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, found-base);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strrchr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	return _wtr_mp_strrchr_impl(vm, ctx, (wt_u8)ch, 0);
}

enum wtr_runresult wtr_mp_strrchr2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;
	wt_u32  offs;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_char, ch);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strrchr_impl(vm, ctx, (wt_u8)ch, offs);
}

static enum wtr_runresult _wtr_mp_strxspn(struct wtr_vm* const vm,
	struct wtr_context* const ctx,
	int (*const cset_func)(const struct cset* const cset, wt_u8 ch),
	wt_u32 offs, wt_vmptr set)
{
	wt_vmptr    str;
	wt_u32      len;
	struct cset cset;

	/* we already checked the pops */
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, set+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	memset(&cset, 0, sizeof(cset));

	set += offsetof(struct wtr_str, data);
	while(0<(len--))
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, set++, &ch, 1))
		{
			goto mav;
		}

		cset_set(&cset, ch);
	}

	/* cset loaded */

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		goto mav;
	}

	if(offs>len)
	{
		ctx->last_exception = IndexOutOfRangeException;
		return WTRR_RAISE;
	}

	str += offsetof(struct wtr_str, data);
	set = str;

	str += offs;
	len -= offs;

	while(0<(len--))
	{
		wt_u8 ch;

		if(0!=wtr_vma_read(vm, str, &ch, 1))
		{
			goto mav;
		}

		if(!cset_func(&cset, ch))
		{
			break;
		}

		++str;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, str-set);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_strspn(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr set;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, set);

	return _wtr_mp_strxspn(vm, ctx, cset_is_set, 0, set);
}

enum wtr_runresult wtr_mp_strspn2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr set;
	wt_u32   offs;

	CHECK_POP(ctx, sizeof(wt_u32)+sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, set);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strxspn(vm, ctx, cset_is_set, offs, set);
}

enum wtr_runresult wtr_mp_strcspn(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr set;

	CHECK_POP(ctx, sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, set);

	return _wtr_mp_strxspn(vm, ctx, cset_is_not_set, 0, set);
}

enum wtr_runresult wtr_mp_strcspn2(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr set;
	wt_u32   offs;

	CHECK_POP(ctx, sizeof(wt_u32)+sizeof(wt_vmptr)*2);
	WTR_CTX_DS_POP(ctx, wt_vmptr, set);
	WTR_CTX_DS_POP(ctx, wt_u32, offs);

	return _wtr_mp_strxspn(vm, ctx, cset_is_not_set, offs, set);
}

enum wtr_runresult wtr_mp_strfromutf8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	((void)vm);

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_strtoutf8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	((void)vm);

	ctx->last_exception = BadProgramAccessException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_chartostring(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char  ch;
	wt_vmptr str;
	char     c;

	CHECK_POP(ctx, sizeof(wt_char));
	CHECK_PUSH(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	c = (char)ch;

	if(0==(str=wtr_str_create_n(vm, &c, 1)))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, str);
}

enum wtr_runresult wtr_mp_charcmp(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_chardiff(vm, ctx);
}

enum wtr_runresult wtr_mp_charadd(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_char rhs;

	CHECK_POP(ctx, sizeof(wt_char)*2);
	WTR_CTX_DS_POP(ctx, wt_char, rhs);
	WTR_CTX_DS_POP(ctx, wt_char, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_char, (lhs+rhs) & 0xff);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charsub(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_char rhs;

	CHECK_POP(ctx, sizeof(wt_char)*2);
	WTR_CTX_DS_POP(ctx, wt_char, rhs);
	WTR_CTX_DS_POP(ctx, wt_char, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_char, (lhs-rhs) & 0xff);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_charinc(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_i32  rhs;

	CHECK_POP(ctx, sizeof(wt_char)+sizeof(wt_i32));
	WTR_CTX_DS_POP(ctx, wt_i32, rhs);
	WTR_CTX_DS_POP(ctx, wt_char, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_char, (lhs+rhs) & 0xff);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chardiff(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char lhs;
	wt_char rhs;

	CHECK_POP(ctx, sizeof(wt_char)*2);
	WTR_CTX_DS_POP(ctx, wt_char, rhs);
	WTR_CTX_DS_POP(ctx, wt_char, lhs);
	WTR_CTX_DS_PUSH(ctx, wt_i32, lhs-rhs);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chartolower(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	if('A'<=((char)ch) && ((char)ch)<='Z')
	{
		ch = ch - 'A' + 'a';
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, ch);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_chartoupper(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char ch;

	CHECK_POP(ctx, sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);

	if('a'<=((char)ch) && ((char)ch)<='z')
	{
		ch = ch - 'a' + 'A';
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, ch);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_seqempty(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr seq;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, seq);

	if(0!=wtr_vma_read(vm, seq+offsetof(struct wtr_seq, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_bool, 0==len);

	return WTRR_RUN;
}

static enum wtr_runresult wtr__mp__cxstring_(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const size_t size,
	char* (*conv_func)(union wtr_genvalue gv, char* buf, char* end))
{
	union wtr_genvalue gv;

	wt_vmptr str;
	char     buf[64];

	CHECK_POP(ctx, size);
	WTR_CTX_STACK_POP(ctx, dsp, size, gv);

	if(0==(str=wtr_str_create_z(vm, conv_func(gv, buf, buf+sizeof(buf)))))
	{
		ctx->last_exception = OutOfMemoryException;
		return WTRR_RAISE;
	}

	return str_push_and_register__tail__(vm, ctx, str);
}

#define _CONVERT_X_TO_STRING(_type_) \
enum wtr_runresult wtr_mp_c##_type_##string(struct wtr_vm* const vm, \
	struct wtr_context* const ctx) \
{ \
	return wtr__mp__cxstring_(vm, ctx, sizeof(wt_##_type_), \
		wtr_convert_##_type_##_to_str); \
}

_CONVERT_X_TO_STRING(i32)
_CONVERT_X_TO_STRING(u32)
_CONVERT_X_TO_STRING(i64)
_CONVERT_X_TO_STRING(u64)
#if WTRUNTIME_ENABLE_FLOAT32
_CONVERT_X_TO_STRING(f32)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
_CONVERT_X_TO_STRING(f64)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
_CONVERT_X_TO_STRING(bool)

static enum wtr_runresult wtr__mp__cstringx_(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const size_t size,
	int (*conv_func)(union wtr_genvalue* const gv, const char* str))
{
	union wtr_genvalue gv;

	wt_vmptr str;
	wt_u32   len;
	char     buf[64];

	CHECK_POP(ctx, sizeof(wt_vmptr));
	CHECK_PUSH(ctx, size);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
	mav:;
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	if((sizeof(buf)-1)<=len) { len = sizeof(buf)-1; }

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, data), buf, len))
	{
		goto mav;
	}

	buf[len] = 0;

	conv_func(&gv, buf);

	if(sizeof(wt_u32)==size)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, gv.v_u32);
	}
	else
	if(sizeof(wt_u64)==size)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u64, gv.v_u64);
	}
	else
	{
		ctx->last_exception = BadProgramAccessException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

#define _CONVERT_STRING_TO_X(_type_) \
enum wtr_runresult wtr_mp_cstring##_type_(struct wtr_vm* const vm, \
	struct wtr_context* const ctx) \
{ \
	return wtr__mp__cstringx_(vm, ctx, sizeof(wt_##_type_), \
		wtr_convert_str_to_##_type_); \
}

_CONVERT_STRING_TO_X(i32)
_CONVERT_STRING_TO_X(u32)
_CONVERT_STRING_TO_X(i64)
_CONVERT_STRING_TO_X(u64)
#if WTRUNTIME_ENABLE_FLOAT32
_CONVERT_STRING_TO_X(f32)
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
#if WTRUNTIME_ENABLE_FLOAT64
_CONVERT_STRING_TO_X(f64)
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
_CONVERT_STRING_TO_X(bool)

enum wtr_runresult wtr_mp_ci32u32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_PEEK(ctx, wt_u32, val);

	if(0x7fffffffU<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32i32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_ci32u32(vm, ctx);
}

enum wtr_runresult wtr_mp_cu32u64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u64, (wt_u64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32i64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_cu32u64(vm, ctx);
}

enum wtr_runresult wtr_mp_ci32u64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i32, val);

	if(0x7fffffffU<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u64, (wt_u64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32i64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_i64));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_i64, (wt_i64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64u32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);

	if(0xffffffffULL<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, (wt_u32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64i32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);

	if(0x7fffffffULL<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, (wt_u32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64u32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_cu64u32(vm, ctx);
}

enum wtr_runresult wtr_mp_ci64i32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i64, val);

	if(0x7fffffffLL<val || (-0x80000000LL)>val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, (wt_i32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64u64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_cu64i64(vm, ctx);
}

enum wtr_runresult wtr_mp_cu64i64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(wt_u64));
	WTR_CTX_DS_PEEK(ctx, wt_u64, val);

	if(0x7fffffffffffffffULL<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_cf32i32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_i32));
	WTR_CTX_DS_POP(ctx, wt_f32, val);

	if((-2147483648.f)>val || (2147483647.f)<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, (wt_i32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf32u32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_f32, val);

	if(0>val || ((wt_f32)0xffffffff)<val)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, (wt_u32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32f32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_f32));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_f32, (wt_f32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32f32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_f32));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_f32, (wt_f32)val);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_ci32f64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_f64));
	WTR_CTX_DS_POP(ctx, wt_i32, val);
	WTR_CTX_DS_PUSH(ctx, wt_f64, (wt_f64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32f64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	CHECK_PUSH(ctx, sizeof(wt_f64));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_f64, (wt_f64)val);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32 && WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_cf32f64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f32 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f32, f);
	WTR_CTX_DS_PUSH(ctx, wt_f64, f);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64f32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f64 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<FLT_MIN || f>FLT_MAX)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_f32, (wt_f32)f);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 && WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_ci64f32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i64, val);
	WTR_CTX_DS_PUSH(ctx, wt_f32, (wt_f32)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64f32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_f32, (wt_f32)val);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_ci64f64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_i64, val);
	WTR_CTX_DS_PUSH(ctx, wt_f64, (wt_f64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64f64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_f64, (wt_f64)val);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64i32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f64 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<-2147483648.0 || f>2147483647.0)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i32, (wt_i32)f);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64u32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f64 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<0 || f>0xffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, (wt_u32)f);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_cf32i64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f32 f;

	CHECK_POP(ctx, sizeof(f));
	CHECK_PUSH(ctx, sizeof(wt_i64));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<(wt_f32)(-0x7fffffffffffffff-1) || f>(wt_f32)0x7fffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i64, (wt_i64)f);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf32u64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f32 f;

	CHECK_POP(ctx, sizeof(f));
	CHECK_PUSH(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<0.f || f>(wt_f32)0xffffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u64, (wt_u64)f);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

#if WTRUNTIME_ENABLE_FLOAT64

enum wtr_runresult wtr_mp_cf64i64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f64 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<(wt_f64)(-0x7fffffffffffffff-1) || f>(wt_f64)0x7fffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_i64, (wt_i64)f);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cf64u64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_f64 f;

	CHECK_POP(ctx, sizeof(f));
	WTR_CTX_DS_POP(ctx, wt_f64, f);

	if(f<0.0 || f>(wt_f64)0xffffffffffffffff)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u64, (wt_u64)f);

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT64 */

enum wtr_runresult wtr_mp_cchari32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_ccharu32(vm, ctx);
}

enum wtr_runresult wtr_mp_ccharu32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char value;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_char, value);
	WTR_CTX_DS_PUSH(ctx, wt_u32, (wt_u32)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cchari64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_ccharu64(vm, ctx);
}

enum wtr_runresult wtr_mp_ccharu64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_char value;

	CHECK_POP(ctx, sizeof(value));
	CHECK_PUSH(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_char, value);
	WTR_CTX_DS_PUSH(ctx, wt_u64, (wt_u64)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32char(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 value;

	CHECK_POP(ctx, sizeof(value));
	WTR_CTX_DS_POP(ctx, wt_i32, value);

	if(255<value || 0>value)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, (wt_char)(wt_u32)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu32char(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 value;

	CHECK_POP(ctx, sizeof(value));
	WTR_CTX_DS_POP(ctx, wt_u32, value);

	if(255<value)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, (wt_char)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci64char(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 value;

	CHECK_POP(ctx, sizeof(value));
	WTR_CTX_DS_POP(ctx, wt_i64, value);

	if(255<value || 0>value)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, (wt_char)(wt_u64)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_cu64char(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 value;

	CHECK_POP(ctx, sizeof(value));
	WTR_CTX_DS_POP(ctx, wt_u64, value);

	if(255<value)
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_char, (wt_char)value);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci32toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i32, u);

	if((-128)<=u && u<=127)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)u) & 0xFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_PEEK(ctx, wt_i32, u);

	if(0<=u && u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32toi16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i32, u);

	if((-32768)<=u && u<=32767)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)u) & 0xFFFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci32tou16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i32 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_PEEK(ctx, wt_i32, u);

	if(0<=u && u<=65535)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32toi16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=32767)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu32tou16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=65535)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i64, u);

	if((-128)<=u && u<=127)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)) & 0xFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i64, u);

	if(0<=u && u<=255)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64toi16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i64, u);

	if((-32768)<=u && u<=32767)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)) & 0xFFFFu);
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci64tou16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_i64 u;

	CHECK_POP(ctx, sizeof(u));
	WTR_CTX_DS_POP(ctx, wt_i64, u);

	if(0<=u && u<=65535)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 u;

	CHECK_POP(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_u64, u);

	if(u<=127)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 u;

	CHECK_POP(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_u64, u);

	if(u<=255)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64toi16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 u;

	CHECK_POP(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_u64, u);

	if(u<=32767)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu64tou16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 u;

	CHECK_POP(ctx, sizeof(wt_u64));
	WTR_CTX_DS_POP(ctx, wt_u64, u);

	if(u<=65535)
	{
		WTR_CTX_DS_PUSH(ctx, wt_u32, ((wt_u32)((wt_u64)u)));
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci8tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu8toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_ci8tou8(vm, ctx);
}

enum wtr_runresult wtr_mp_ci16toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, u);

	if(u<=127)
	{
		/* ok */
	}
	else
	if(0xFF80<u && u<=0xFFFF)
	{
		u &= 0xFF;
	}
	else
	{
		ctx->last_exception = ConstraintErrorException;
		return WTRR_RAISE;
	}

	WTR_CTX_DS_PUSH(ctx, wt_u32, u);
	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_ci16tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_ci16tou16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=0x7FFF)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu16toi16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_ci16tou16(vm, ctx);
}

enum wtr_runresult wtr_mp_cu16toi8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=127)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_cu16tou8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 u;

	CHECK_POP(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PEEK(ctx, wt_u32, u);

	if(u<=255)
	{
		return WTRR_RUN;
	}

	ctx->last_exception = ConstraintErrorException;
	return WTRR_RAISE;
}

enum wtr_runresult wtr_mp_version(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, WTR_VERSION);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_compversion(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, WTR_BINCOMPSINCEVER);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_coreversion(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return wtr_mp_version(vm, ctx);
}

enum wtr_runresult wtr_mp_nullhandle(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	CHECK_PUSH(ctx, sizeof(wt_handle));
	WTR_CTX_DS_PUSH(ctx, wt_handle, 0);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_handleequal(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle lhs;
	wt_handle rhs;

	CHECK_POP(ctx, sizeof(wt_handle)*2);
	WTR_CTX_DS_POP(ctx, wt_handle, lhs);
	WTR_CTX_DS_POP(ctx, wt_handle, rhs);
	WTR_CTX_DS_PUSH(ctx, wt_bool, lhs==rhs);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_fileaccess(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_i32   mode;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_i32));
	WTR_CTX_DS_POP(ctx, wt_i32, mode);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);
	ASSERT_OBJECT_EXISTENCE(str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_access(ctx, str+offsetof(struct wtr_str, data), len, mode);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filestat(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	CHECK_PUSH(ctx, sizeof(wt_bool)+sizeof(wt_char)+sizeof(wt_u64)*3);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);
	ASSERT_OBJECT_EXISTENCE(str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_stat(ctx, str+offsetof(struct wtr_str, data), len);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filestatf(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	CHECK_PUSH(ctx, sizeof(wt_bool)+sizeof(wt_char)+sizeof(wt_u64)*3);
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_fstat(ctx, (wt_file)(wt_uintptr)han);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileopen(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;
	wt_char  ch;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, ch);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	ASSERT_OBJECT_EXISTENCE(str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_fopen(ctx, str+offsetof(struct wtr_str, data), len, (char)ch);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileclose(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_fclose(ctx, (wt_file)(wt_uintptr)han);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileeof(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_feof(ctx, (wt_file)(wt_uintptr)han);
	return WTRR_IOCALL;
}

static enum wtr_runresult _wtr_mp_fileread_impl(struct wtr_vm* const vm,
	struct wtr_context* const ctx, const int pop_and_check_return_value)
{
	wt_handle han;
	wt_vmptr  addr;
	wt_u32    size;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_vmptr)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, size);
	WTR_CTX_DS_POP(ctx, wt_vmptr, addr);
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_fread(ctx, (wt_file)(wt_uintptr)han, addr, size,
		pop_and_check_return_value);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_fileread(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_fileread_impl(vm, ctx, 1);
}

enum wtr_runresult wtr_mp_filereadr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp_fileread_impl(vm, ctx, 0);
}

enum wtr_runresult wtr_mp_filereadstr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;
	wt_char   sep;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_char));
	WTR_CTX_DS_POP(ctx, wt_char, sep);
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_freadln(ctx, (wt_file)(wt_uintptr)han, ctx->sbuf,
		ctx->sbuf_size, (char)sep);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filewrite(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;
	wt_vmptr  addr;
	wt_u32    size;

	CHECK_POP(ctx, sizeof(wt_handle)+sizeof(wt_vmptr)+sizeof(wt_u32));
	WTR_CTX_DS_POP(ctx, wt_u32, size);
	WTR_CTX_DS_POP(ctx, wt_vmptr, addr);
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_fwrite(ctx, (wt_file)(wt_uintptr)han, addr, size, 1);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_filewritestr(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;
	wt_vmptr  str;
	wt_u32    len;
	wt_char   sep;

	CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_char)+sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_char, sep);
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_fwriteln(ctx, (wt_file)(wt_uintptr)han,
		str+offsetof(struct wtr_str, data), len, (char)sep);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_diropen(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_vmptr str;
	wt_u32   len;

	CHECK_POP(ctx, sizeof(wt_vmptr));
	WTR_CTX_DS_POP(ctx, wt_vmptr, str);

	if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length), &len,
		sizeof(len)))
	{
		ctx->last_exception = MemoryAccessViolationException;
		return WTRR_RAISE;
	}

	wtr_iocall_opendir(ctx, str+offsetof(struct wtr_str, data), len);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirclose(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_closedir(ctx, (wt_dir)(wt_uintptr)han);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirread(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_readdir(ctx, (wt_dir)(wt_uintptr)han);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirgetname(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_dirgetname(ctx, (wt_dir)(wt_uintptr)han, ctx->sbuf,
		ctx->sbuf_size);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_dirgettype(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_handle han;

	CHECK_POP(ctx, sizeof(wt_handle));
	WTR_CTX_DS_POP(ctx, wt_handle, han);

	wtr_iocall_dirgettype(ctx, (wt_dir)(wt_uintptr)han);
	return WTRR_IOCALL;
}

enum wtr_runresult wtr_mp_rand(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 num = _wtr_xorshift(vm);

	((void)vm);

	CHECK_PUSH(ctx, sizeof(wt_u32));
	WTR_CTX_DS_PUSH(ctx, wt_u32, num);

	return WTRR_RUN;
}

#if WTRUNTIME_ENABLE_FLOAT32

enum wtr_runresult wtr_mp_randf(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 num = _wtr_xorshift(vm);

	((void)vm);

	CHECK_PUSH(ctx, sizeof(wt_f32));
	WTR_CTX_DS_PUSH(ctx, wt_f32, ((float)(num&0x00FFFFFF)) / ((float)0x00FFFFFF));

	return WTRR_RUN;
}

#endif /* WTRUNTIME_ENABLE_FLOAT32 */

enum wtr_runresult wtr_mp_srand(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 seed;

	((void)vm);

	CHECK_POP(ctx, sizeof(seed));
	WTR_CTX_DS_POP(ctx, wt_u32, seed);

	vm->xs_x = (seed);
	vm->xs_y = (seed >> 13);
	vm->xs_z = (seed << 7);
	vm->xs_w = (seed ^ 0xa7a7a7a7);

	_wtr_xorshift(vm);

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_islittleendian(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	CHECK_PUSH(ctx, sizeof(wt_bool));
	WTR_CTX_DS_PUSH(ctx, wt_bool, WTBO_LITTLE==wtr_byteorder());

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_isbigendian(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	CHECK_PUSH(ctx, sizeof(wt_bool));
	WTR_CTX_DS_PUSH(ctx, wt_bool, WTBO_BIG==wtr_byteorder());

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_swap16((wt_u16)val));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_swap32(val));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_swap64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_u64, wtr_swap64(val));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_botoh16((wt_u16)val, WTBO_LITTLE));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_botoh32(val, WTBO_LITTLE));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_lswap64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_u64, wtr_botoh64(val, WTBO_LITTLE));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_botoh16((wt_u16)val, WTBO_BIG));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u32 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u32, val);
	WTR_CTX_DS_PUSH(ctx, wt_u32, wtr_botoh32(val, WTBO_BIG));

	return WTRR_RUN;
}

enum wtr_runresult wtr_mp_bswap64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	wt_u64 val;

	CHECK_POP(ctx, sizeof(val));
	WTR_CTX_DS_POP(ctx, wt_u64, val);
	WTR_CTX_DS_PUSH(ctx, wt_u64, wtr_botoh64(val, WTBO_BIG));

	return WTRR_RUN;
}

static enum wtr_runresult wtr_mp__chrclassify_impl(
	struct wtr_vm* const vm,
	struct wtr_context* const ctx, int (*chrclassify)(int))
{
	wt_u32 ch;

	CHECK_POP(ctx, sizeof(ch));
	WTR_CTX_DS_POP(ctx, wt_u32, ch);
	WTR_CTX_DS_PUSH(ctx, wt_bool, (wt_bool)chrclassify((char)ch));

	return WTRR_RUN;
}

#define _DEF_CHAR_CLASSIFICATION_FUNC(_name_) \
enum wtr_runresult wtr_mp_##_name_(struct wtr_vm* const vm, struct wtr_context* const ctx) \
{ \
	return wtr_mp__chrclassify_impl(vm, ctx, _name_); \
}

_DEF_CHAR_CLASSIFICATION_FUNC(iscntrl)
_DEF_CHAR_CLASSIFICATION_FUNC(isprint)
_DEF_CHAR_CLASSIFICATION_FUNC(isspace)
_DEF_CHAR_CLASSIFICATION_FUNC(isblank)
_DEF_CHAR_CLASSIFICATION_FUNC(isgraph)
_DEF_CHAR_CLASSIFICATION_FUNC(ispunct)
_DEF_CHAR_CLASSIFICATION_FUNC(isalnum)
_DEF_CHAR_CLASSIFICATION_FUNC(isalpha)
_DEF_CHAR_CLASSIFICATION_FUNC(isupper)
_DEF_CHAR_CLASSIFICATION_FUNC(islower)
_DEF_CHAR_CLASSIFICATION_FUNC(isdigit)
_DEF_CHAR_CLASSIFICATION_FUNC(isxdigit)

static wt_u32 wtr_clz32(wt_u32 x)
{
	wt_u32 n = 1;

	if(0==x) { return 32; }
	if(0==(x>>16)) { n += 16; x <<= 16; }
	if(0==(x>>24)) { n +=  8; x <<=  8; }
	if(0==(x>>28)) { n +=  4; x <<=  4; }
	if(0==(x>>30)) { n +=  2; x <<=  2; }
	n -= (x>>31);

	return n;
}

static wt_u32 wtr_ctz32(wt_u32 x)
{
	wt_u32 n = 1;

	if(0==x) { return 32; }
	if(0==(x & (wt_u32)(0x0000FFFFu))) { n += 16; x >>= 16;}
	if(0==(x & (wt_u32)(0x000000FFu))) { n +=  8; x >>=  8;}
	if(0==(x & (wt_u32)(0x0000000Fu))) { n +=  4; x >>=  4;}
	if(0==(x & (wt_u32)(0x00000003u))) { n +=  2; x >>=  2;}

	return n - (x & 1);
}

static wt_u64 wtr_clz64(wt_u64 x)
{
	wt_u64 n = 1;

	if(0==x) { return 64; }
	if(0==(x>>32)) { n += 32; x <<= 32; }
	if(0==(x>>48)) { n += 16; x <<= 16; }
	if(0==(x>>56)) { n +=  8; x <<=  8; }
	if(0==(x>>60)) { n +=  4; x <<=  4; }
	if(0==(x>>62)) { n +=  2; x <<=  2; }
	n -= (x>>63);

	return n;
}

static wt_u64 wtr_ctz64(wt_u64 x)
{
	wt_u64 n = 1;

	if(0==x) { return 64; }
	if(0==(x & (wt_u64)(0xFFFFFFFFul))) { n += 32; x >>= 32;}
	if(0==(x & (wt_u64)(0x0000FFFFul))) { n += 16; x >>= 16;}
	if(0==(x & (wt_u64)(0x000000FFul))) { n +=  8; x >>=  8;}
	if(0==(x & (wt_u64)(0x0000000Ful))) { n +=  4; x >>=  4;}
	if(0==(x & (wt_u64)(0x00000003ul))) { n +=  2; x >>=  2;}

	return n - (wt_u64)(x & 1);
}

static wt_u32 _wrap_clz8(wt_u32 val)
{
	return wtr_clz32(val & 0xFF) - 24;
}

enum wtr_runresult wtr_mp_clz8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_clz8);
}

static wt_u32 _wrap_clz16(wt_u32 val)
{
	return wtr_clz32(val & 0xFFFF) - 16;
}

enum wtr_runresult wtr_mp_clz16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_clz16);
}

enum wtr_runresult wtr_mp_clz32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, wtr_clz32);
}

enum wtr_runresult wtr_mp_clz64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u64(vm, ctx, wtr_clz64);
}

static wt_u32 _wrap_ctz8(wt_u32 val)
{
	return wtr_ctz32(val | 0x100);
}

enum wtr_runresult wtr_mp_ctz8(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_ctz8);
}

static wt_u32 _wrap_ctz16(wt_u32 val)
{
	return wtr_ctz32(val | 0x10000);
}

enum wtr_runresult wtr_mp_ctz16(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, _wrap_ctz16);
}

enum wtr_runresult wtr_mp_ctz32(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u32(vm, ctx, wtr_ctz32);
}

enum wtr_runresult wtr_mp_ctz64(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	return _wtr_mp__g_u_u64(vm, ctx, wtr_ctz64);
}
